type graphe = int list array 

let g0 = [|
	[1];
	[2; 3];
	[0; 3];
	[]
|]

(* nombre d'arêtes du graphe orienté g *)
let nb_aretes (g: graphe) : int =
	let n = Array.length g in 
	let res = ref 0 in 
	for i = 0 to n-1 do
		res := !res + List.length g.(i)
	done;
	!res


(* nombre d'arêtes du graphe non-orienté g (supposé sans boucles) *)
let nb_aretes_no (g: graphe) : int = 
	nb_aretes g / 2


(* a puissance b (pour b >= 0) *)
let fast_exp (a: int) (b: int) =
	let aa = ref a in 
	let bb = ref b in 
	let r = ref 1 in
	while !bb > 0 do 
		if !bb mod 2 = 1 then 
			r := !r * !aa;
		aa := !aa * !aa;
		bb := !bb / 2
	done;
	!r



type 'a pile = 'a list ref

(* crée une pile vide *)
let pile_vide () : 'a pile = ref []

(* empile x sur a *)
let empiler (p: 'a pile) (x: 'a) : unit = 
	p := x :: !p

(* dépile le sommet de p (supposée non vide) et le renvoie *)
let depiler (p: 'a pile) : 'a =
	match !p with 
	| [] -> failwith "pile vide"
	| x :: q -> p := q; x 

(* pour chaque sommet i de la composante connexe de s dans g, écrit k dans
   la case c.(i).
   Précondition: les cases de c correspondant à la composante connexe de s
   contiennent toutes des -1 *)
let noter_composante_connexe_source (g: graphe) (s: int) (k: int) (c: int array) =
	let p = pile_vide () in
	empiler p s;
	c.(s) <- k;
	while !p <> [] do
		let u = depiler p in
		(* transformer la boucle for du parcours en une fonction *)
		let rec traiter_voisins (l: int list) : unit =
			match l with 
			| [] -> ()
			| v :: q -> begin 
				if c.(v) = -1 then begin
					empiler p v;
					c.(v) <- k
				end; traiter_voisins q
			end
		in traiter_voisins g.(u)
	done

(* renvoie un tableau c associant à chaque sommet de g l'indice de sa 
   composante connexe. Deux sommets u, v de g tels que c.(u) = c.(v)
   sont dans la même composante connexe. *)
let composantes_connexes (g: graphe) : int array = 
	let n = Array.length g in
	let c = Array.make n (-1) in
	let k = ref 1 in 
	for s = 0 to n-1 do
		if c.(s) = -1 then begin 
			noter_composante_connexe_source g s !k c;
			k := !k +1
		end
	done; c